### this file calls the vocola Keys extension, http://vocola.net/unofficial/keys.html
### this software is released under the MIT license.
"""Sending keystrokes to the foreground window.

  Usage:

At top your module insert:

:code:`from dtactions.sendkeys import sendkeys`

and then in a function:
    
    :code:`sendkeys("keystrokes")`
    
Optionally, you can also use sendsystemkeys, which is implmented via Dragon SendSystemKeys (via natlink.execScript)

(Quintijn Hoogenboom, 2021-04-04)
"""

import natlink
from dtactions.vocola_sendkeys import ext_keys

def sendkeys(keys):
    """sends keystrokes via the vocola Keys extension 
   
:code:`"{shift+right 4}"`

This functions is similar to the Dragonfly `sendkeys` function, but now works via the Vocola Keys extension.

Tested at bottom of this file interactively...
    """
    ext_keys.send_input(keys)
    
def sendsystemkeys(keys):
    """sends keystrokes "the hard way" via Dragon's SendSystemKeys
   
Tested at bottom of this file interactively...
    """
    if not keys:
        return
    natlink.execScript(f'SendSystemKeys("{keys}")')
    
    
    
if __name__ == "__main__":
    # sendkeys("{a 3}") #aaa
    # sendkeys("x y z ")
    # sendkeys("test, test, met komma.{home}")
    # try:
    #     natlink.natConnect()
    #     sendsystemkeys("a{win+e}b")
    # finally:
    #     natlink.natDisconnect()
    
    # sendkeys("{ctrl+end}{up 2}{home}{shift+end}{del}this is wrong{shift+left 5}right")
    # ##

    """
    """



"""utility functions from Quintijn, used in unimacro and in local programs.
   in python3 also the path module as subclass of the standard path class
"""
#pylint:disable=C0116, C0302,  W0613, R0912, R0914, R0915, R0911, W0702
import unicodedata
import os
import re
import collections
from pathlib import Path
    
def unifyaccentedchars(to_translate):
    """change acuted characters with combining code to single characters
    
The combining variant (s3) is converted into the one character shorter variant s1, apart from the capitals.
This is done by the NFC variant of the normalize function
    
>>> s1 = "cafcomb\u00e9"   # single char e acute no change
>>> s2 = unifyaccentedchars(s1)
>>> s2
'cafcomb\u00e9'

## this one changes to s1:
>>> s3 = "CafCombe\u0301"   # combining char e acute   0301 is the combining code
>>> s4 = unifyaccentedchars(s3)   # combining char e acute
>>> s4
'CafComb\u00e9'
>>> len(s1), len(s2), len(s3), len(s4)
(8, 8, 9, 8)
>>> s1 == s3.lower()
False
>>> s2 == s4.lower()
True

    (from Fluent Python)
    """
    norm_txt = unicodedata.normalize('NFC', to_translate)
    return norm_txt

def normalizeaccentedchars(to_translate):
    """change acutechars to ascii 
    
>>> s1 = "cafnon\u00e9"   # single char e acute
>>> normalizeaccentedchars(s1)  
'cafnone'
>>> s2 = "CafCombe\u0301"   # combining char e acute   0301 is the combining code
>>> normalizeaccentedchars(s2)   # combining char e acute
'CafCombe'
>>> len(s1), len(s2)
(7, 9)

    (from Fluent Python)
    """
    norm_txt = unicodedata.normalize('NFD', to_translate)
    shaved = ''.join(c for c in norm_txt if not unicodedata.combining(c))
    return shaved

def doubleaccentedchars(to_translate):
    """change acutechars to ascii, but double them e acute = ee
    (from Fluent Python, adaptation QH)
    

>>> s1 = "double caf\\u00e9"   # single char e acute
>>> doubleaccentedchars(s1)
'double cafee'
>>> s2 = "Double Cafe\\u0301"   # combining char e acute
>>> doubleaccentedchars(s2)
'Double Cafee'

>>> doubleaccentedchars("enqu\N{LATIN SMALL LETTER E}\N{COMBINING CIRCUMFLEX ACCENT}te")
'enqueete'

    """
    norm_txt = unicodedata.normalize('NFD', to_translate) ## haal char en accent uit elkaar
    shaved = []
    last = ""
    for c in norm_txt:
        comb = unicodedata.combining(c)
        if comb:
            if comb == 230:
                # accent aegu, accent grave, accent circonflex, decide in favour of accent aegu, double char
                # print('doubleaccentedchars, combining value %s, double char: %s (%s)'% (comb, last, to_translate))
                shaved.append(last)
            elif comb == 202:
                if last == "c":
                    # print('c cedilla, change to "s" (%s)'% to_translate)
                    shaved.pop()
                    shaved.append("s")
                else:
                    # print("c cedilla, but NO C, ignore (%s)"% to_translate)
                    pass
            else:
                print(f'(yet) unknown combining char {comb} in "{to_translate}", ignore')
            last = ""
        else:
            shaved.append(c)
            last = c
    return ''.join(shaved)

class peek_ahead:
    """ An iterator that supports a peek operation.
    
    Improved for python3 after QH's example: Adapted for python 3 by Paulo Roma.
    
    this is a merge of example 19.18 of python cookbook part 2, peek ahead more steps
    and the simpler example 16.7, which peeks ahead one step and stores it in
    the self.preview variable.
    
    Adapted so the peek function never raises an error, but gives the
    self.sentinel value in order to identify the exhaustion of the iter object.
    
    Example usage (Paulo):
    
    >>> p = peek_ahead(range(4))
    >>> p.peek()
    0
    >>> p.next(1)
    [0]
    >>> p.isFirst()
    True
    >>> p.preview
    1
    >>> p.isFirst()
    True
    >>> p.peek(3)
    [1, 2, 3]
    >>> p.next(2)
    [1, 2]
    >>> p.peek(2) #doctest: +ELLIPSIS
    [3, <object object at ...>]
    >>> p.peek(1)
    [3]
    >>> p.next(2)
    Traceback (most recent call last):
    StopIteration
    >>> p.next()
    3
    >>> p.isLast()
    True
    >>> p.next()
    Traceback (most recent call last):
    StopIteration
    >>> p.next(0)
    []
    >>> p.peek()  #doctest: +ELLIPSIS
    <object object at ...>
    >>> p.preview #doctest: +ELLIPSIS
    <object object at ...>
    >>> p.isLast()  # after the iter process p.isLast remains True
    True
    
    ### my old unittests, QH:
    
        >>> p = peek_ahead(range(4))
    >>> p.peek()
    0
    >>> p.next(1)
    [0]
    >>> p.isFirst()
    True
    >>> p.preview
    1
    >>> p.isFirst()
    True
    >>> p.peek(3)
    [1, 2, 3]
    >>> p.next(2)
    [1, 2]
    >>> p.peek(2) #doctest: +ELLIPSIS
    [3, <object object at ...>]
    >>> p.peek(4) #doctest: +ELLIPSIS
    [3, <object object at ...>, <object object at ...>, <object object at ...>]
    >>> p.peek(1)
    [3]
    >>> p.next(2)
    Traceback (most recent call last):
    StopIteration
    >>> p.next()
    3
    >>> p.isLast()
    True
    >>> p.next()
    Traceback (most recent call last):
    StopIteration
    >>> p.next(0)
    []
    >>> p.peek()  #doctest: +ELLIPSIS
    <object object at ...>
    >>> p.preview #doctest: +ELLIPSIS
    <object object at ...>
    >>> p.isLast()  # after the iter process p.isLast remains True
    True

    From example 16.7 from python cookbook 2.

    The preview can be inspected through it.preview

    ignoring duplicates:
    >>> it = peek_ahead('122345567')
    >>> for i in it:
    ...     if it.preview == i:
    ...         continue
    ...     print(i, end=" ")
    1 2 3 4 5 6 7 

    getting duplicates together:
    >>> it = peek_ahead('abbcdddde')
    >>> for i in it:
    ...     if it.preview == i:
    ...         dup = 1
    ...         while 1:
    ...             i = it.next()
    ...             dup += 1
    ...             if i != it.preview:
    ...                 print(i*dup, end=" ")
    ...                 break
    ...     else:
    ...         print(i, end=" ")
    ...
    a bb c dddd e 
    
    """
    ## schildwacht (guard)
    sentinel = object()
    def __init__(self, iterable):
        ## iterator
        self._iterable = iter(iterable)
        try:
           ## next method hold for speed
            self._nit = self._iterable.next
        except AttributeError:
            self._nit = self._iterable.__next__
        ## deque object initialized left-to-right (using append())
        self._cache = collections.deque()
        ## initialize the first preview already
        self._fillcache(1)
        ## peek at leftmost item
        self.preview = self._cache[0]
        ## keeping the count allows checking isFirst and isLast status
        self.count = -1

    def __iter__(self):
        """return an iterator
        """
        return self

    def _fillcache(self, n):
        """fill _cache of items to come, with one extra for the preview variable
        """
        if n is None:
            n = 1
        while len(self._cache) < n+1:
            try:
                Next = self._nit()
            except StopIteration:
                # store sentinel, to identify end of iter:
                Next = self.sentinel
            self._cache.append(Next)

    def __next__(self, n=None):
        """gives next item of the iter, or a list of n items
        
        raises StopIteration if the iter is exhausted (self.sentinel is found),
        but in case of n > 1 keeps the iter alive for a smaller "next" calls
        """
        self._fillcache(n)
        if n is None:
            result = self._cache.popleft()
            if result == self.sentinel:
                # find sentinel, so end of iter:
                self.preview = self._cache[0]
                raise StopIteration
            self.count += 1
        else:
            result = [self._cache.popleft() for i in range(n)]
            if result and result[-1] == self.sentinel:
                # recache for future use:
                self._cache.clear()
                self._cache.extend(result)
                self.preview = self._cache[0]
                raise StopIteration
            self.count += n
        self.preview = self._cache[0]
        return result

    def next(self,n=None):
        """python2 compatibility
        """
        #pylint:disable=C2801
        return self.__next__(n)
    
    def isFirst(self):
        """returns true if iter is at first position
        """
        return self.count == 0

    def isLast(self):
        """returns true if iter is at last position or after StopIteration
        """
        return self.preview == self.sentinel

    def hasNext(self):
        """returns true if iter is not at last position
        """
        return not self.isLast()
        
    def peek(self, n=None):
        """gives next item, without exhausting the iter, or a list of 0 or more next items
        
        with n == None, you can also use the self.preview variable, which is the first item
        to come.
        """
        self._fillcache(n)
        if n is None:
            result = self._cache[0]
        else:
            result = [self._cache[i] for i in range(n)]
        return result
    # old name:
    # peekable = peek_ahead

## TODOQH
class peek_ahead_stripped(peek_ahead):
    """ Iterator that strips lines of text, and returns (leftSpaces,strippedLine)

    sentinel is just False, such that peeking ahead can check for truth input

    >>> lines = ['line1', '', ' one space ahead','', '   three spaces ahead, 1 empty line before']
    >>> list(peek_ahead_stripped(lines))
    [(0, 'line1'), (0, ''), (1, 'one space ahead'), (0, ''), (3, 'three spaces ahead, 1 empty line before')]

    example of testing look ahead

    >>> lines = ['line1 ', '', 'line2 (last)']
    >>> it = peek_ahead_stripped(lines)
    >>> for spaces, text in it:
    ...     print('current line: |', text, '|', end=' ')
    ...     if it.preview is it.sentinel:
    ...         print(', cannot preview, end of peek_ahead_stripped')
    ...     elif it.preview[1]:
    ...         print(', non empty preview: |', it.preview[1], '|')
    ...     else:
    ...         print(', empty preview')
    current line: | line1 | , empty preview
    current line: |  | , non empty preview: | line2 (last) |
    current line: | line2 (last) | , cannot preview, end of peek_ahead_stripped

    """
    def _fillcache(self, n):
        """fill _cache of items to come, special treatment for this stripped subclass
        """
        if n is None:
            n = 1
        while len(self._cache) < n+1:
            try:
                line = self._nit()
                line = line.rstrip()
                Next = (len(line) - len(line.lstrip()), line.lstrip())
            except StopIteration:
                # store sentinel, to identify end of iter:
                Next = self.sentinel
            self._cache.append(Next)
     
# helper string functions:
def replaceExt(fileName, ext):
    """change extension of file

>>> replaceExt("a.psd", ".jpg")
'a.jpg'
>>> replaceExt("a/b/c/d.psd", "jpg")
'a/b/c/d.jpg'
    """
    ext = addToStart(ext, ".")
    fileName = str(fileName)
    a, _extOld = os.path.splitext(fileName)
    return a + ext

def getExt(fileName):
    """return the extension of a file

>>> getExt(u"a.psd")
'.psd'
>>> getExt("a/b/c/d.psd")
'.psd'
>>> getExt("abcd")
''
>>> getExt("a/b/xyz")
''
    """
    _a, ext = os.path.splitext(fileName)
    return ext

def removeFromStart(text, toRemove, ignoreCase=None):
    """returns the text with "toRemove" stripped from the start if it matches
>>> removeFromStart('abcd', 'a')
'bcd'
>>> removeFromStart('abcd', 'not')
'abcd'

working of ignoreCase:

>>> removeFromStart('ABCD', 'a')
'ABCD'
>>> removeFromStart('ABCD', 'ab', ignoreCase=1)
'CD'
>>> removeFromStart('abcd', 'ABC', ignoreCase=1)
'd'

    """
    if ignoreCase:
        text2 = text.lower()
        toRemove = toRemove.lower()
    else:
        text2 = text
    if text2.startswith(toRemove):
        return text[len(toRemove):]
    return text


def removeFromEnd(text, toRemove, ignoreCase=None):
    """returns the text with "toRemove" stripped from the end if it matches

>>> removeFromEnd('a.jpg', '.jpg')
'a'
>>> removeFromEnd('b.jpg', '.gif')
'b.jpg'

working of ignoreCase:

>>> removeFromEnd('C.JPG', '.jpg')
'C.JPG'
>>> removeFromEnd('D.JPG', '.jpg', ignoreCase=1)
'D'
>>> removeFromEnd('d.jpg', '.JPG', ignoreCase=1)
'd'

    """
    if ignoreCase:
        text2 = text.lower()
        toRemove = toRemove.lower()
    else:
        text2 = text
    if text2.endswith(toRemove):
        return text[:-len(toRemove)]
    return text

def addToStart(text, toAdd, ignoreCase=None):
    """returns text with "toAdd" added at the start if it was not already there

    if ignoreCase:
        return the start of the string with the case as in "toAdd"

>>> addToStart('a-text', 'a-')
'a-text'
>>> addToStart('text', 'b-')
'b-text'
>>> addToStart('B-text', 'b-')
'b-B-text'

working of ignoreCase:

>>> addToStart('C-Text', 'c-', ignoreCase=1)
'c-Text'
>>> addToStart('d-Text', 'D-', ignoreCase=1)
'D-Text'

    """
    if ignoreCase:
        text2 = text.lower()
        toAdd2 = toAdd.lower()
    else:
        text2 = text
        toAdd2 = toAdd
    if text2.startswith(toAdd2):
        return toAdd + text[len(toAdd):]
    return toAdd + text

def addToEnd(text, toAdd, ignoreCase=None):
    """returns text with "toAdd" added at the end if it was not already there

    if ignoreCase:
        return the end of the string with the case as in "toAdd"

>>> addToEnd('a.jpg', '.jpg')
'a.jpg'
>>> addToEnd('b', '.jpg')
'b.jpg'

working of ignoreCase:

>>> addToEnd('Cd.JPG', '.jpg', ignoreCase=1)
'Cd.jpg'
>>> addToEnd('Ef.jpg', '.JPG', ignoreCase=1)
'Ef.JPG'

    """
    if ignoreCase:
        text2 = text.lower()
        toAdd2 = toAdd.lower()
    else:
        text2 = text
        toAdd2 = toAdd
    if text2.endswith(toAdd2):
        return text[:-len(toAdd)] + toAdd
    return text + toAdd

def firstLetterCapitalize(t):
    """capitalize only the first letter of the string
    """
    if t:
        return t[0].upper() + t[1:]
    return ""

def extToLower(fileName):
    """leave name part intact, but change extension to lowercase
>>> extToLower("aBc.jpg")
'aBc.jpg'
>>> extToLower("ABC.JPG")
'ABC.jpg'
>>> extToLower("D:/a/B/ABC.JPG")
'D:/a/B/ABC.jpg'



    """
    f, ext = os.path.splitext(fileName)
    return f + ext.lower()


def appendBeforeExt(text, toAppend):
    """append text just before the extension of the filename part

>>> appendBeforeExt("short.html", '__t')
'short__t.html'
>>> appendBeforeExt("http://a/b/c/d/long.html", '__b')
'http://a/b/c/d/long__b.html'
    """
    base, ext = os.path.splitext(text)
    return base + toAppend + ext

Classes = ('ExploreWClass', 'CabinetWClass')


## to pathqh:
# # def toUnixName(t, glueChar="", lowercase=1, canHaveExtension=1, canHaveFolders=1, mayBeEmpty=False):
    
def sendkeys_escape(Str):
    """escape with {} keys that have a special meaning in sendkeys
    + ^ % ~ { } [ ]

>>> sendkeys_escape('abcd')
'abcd'
>>> sendkeys_escape('+bcd')
'{+}bcd'
>>> sendkeys_escape('+a^b%c~d{f}g[h]i')
'{+}a{^}b{%}c{~}d{{}f{}}g{[}h{]}i'
>>> sendkeys_escape('+^%~{}[]')
'{+}{^}{%}{~}{{}{}}{[}{]}'

    """
    ## make str, for python 2 for the time being:
    return ''.join(map(_sendkeys_escape, Str))

def _sendkeys_escape(s):
    """Escape one character in the set or return if different"""
    if s in ('+', '^', '%', '~', '{' , '}' , '[' , ']' ) :
        result = '{%s}' % s
    else:
        result = s
    return str(result)

def getSublists(L, maxLen, sepLen):
    """generator function, that gives pieces of the list, up to
    the maximum length, accounting for the separator length
    """
    if not L:
        yield L
        return
    listPart = [L[0]]
    lenPart = len(L[0])
    for w in L[1:]:
        lw = len(w)
        if lw + lenPart > maxLen:
            yield listPart
            listPart = [w]
            lenPart = lw
        else:
            lenPart += lw + sepLen
            listPart.append(w)
    yield listPart

def getWordsUntilLength(t, maxLength):
    """take words until maxLength is reached
>>> getWordsUntilLength('this is a test', 60)
'this is a test'
>>> getWordsUntilLength('this is a test', 7)
'this is'
>>> getWordsUntilLength('this is a test', 2)
'this'
    
    
    """
    t = t.replace(',', '')
    t = t.replace('.', '')
    T = t.split()
    while T:
        t = ' '.join(T)
        if len(t) <= maxLength:
            return t
        T.pop()
    return t

def splitLongString(S, maxLen=70, prefix='', prefixOnlyFirstLine=0):
    """Splits a (long) string into newline separated parts,

    a list of strings is returned.

    possibly with a fixed prefix, or a prefix for the first line only.
    Possibly items inside the line are separated by a given separator

    maxLen = maximum line length, can be exceeded is a very long word is there
    prefix = text that is inserted in front of each line, default ''
    prefixOnlyFirstLine = 1: following lines as blank prefix, default 0
    >>> splitLongString('foo', 80)
    ['foo']
    >>> splitLongString(' foo   bar and another set of  words  ', 80)
    ['foo bar and another set of words']
    >>> splitLongString(' foo   bar and another set of  words  ', 20,
    ... prefix='    # ')
    ['    # foo bar and', '    # another set of', '    # words']
    >>> splitLongString(' foo   bar and another set of  words  ', 20,
    ... prefix='entry = ', prefixOnlyFirstLine=1)
    ['entry = foo bar and', '        another set', '        of words']
    """
    assert isinstance(S, str)
    L = [t.strip() for t in S.split()]
    lOut = []
    for part in getSublists(L, maxLen=maxLen-len(prefix), sepLen=1):
        lOut.append(prefix + ' '.join(part))
        if prefixOnlyFirstLine:
            prefix = ' '*len(prefix)
    return lOut



def cleanString(s):
    """converts a string with leading and trailing and
    intermittent whitespace into a string that is stripped
    and has only single spaces between words
>>> cleanString('foo bar')
'foo bar'
>>> cleanString('foo  bar')
'foo bar'
>>> cleanString('\\n foo \\n\\n  bar ')
'foo bar'
>>> cleanString('')
''

    """
    return ' '.join([x.strip() for x in s.split()])

def formatListColumns(List, lineLen = 70, sort = 0):
    """formats a list in columns

    Uses a generator function "splitList", that gives a sequence of
    sub lists of length n.

    The items are separated by at least two spaces, if the list
    can be placed on one line, the list is comma separated

>>> formatListColumns([''])
''
>>> formatListColumns(['a','b'])
'a, b'
>>> formatListColumns(['foo', 'bar', 'longer entry'], lineLen=5)
'foo\\nbar\\nlonger entry'
>>> formatListColumns(['foo', 'bar', 'longer entry'], lineLen=5, sort=1)
'bar\\nfoo\\nlonger entry'
>>> print(formatListColumns(['afoo', 'bar', 'clonger', 'dmore', 'else', 'ftest'], lineLen=20, sort=1))
afoo     dmore
bar      else
clonger  ftest
>>> print(formatListColumns(['foo', 'bar', 'longer entry'], lineLen=20))
foo  longer entry
bar

    """
    if sort:
        List = sorted(List, key=str.casefold)
    s = ', '.join(List)

    # short list, simply join with comma space:
    if len(s) <= lineLen:
        return s

    maxLen = max(list(map(len, List)))

    # too long elements in list, return "\n" separated string:
    if maxLen > lineLen:
        return '\n'.join(List)


    nRow = len(s)//lineLen + 1
    lenList = len(List)

    # try for successive number of rows:
    while nRow < lenList//2 + 2:
        lines = []
        for i in range(nRow):
            lines.append([])
        maxLenTotal = 0
        for parts in splitList(List, nRow):
            maxLenParts = max(list(map(len, parts))) + 2
            maxLenTotal += maxLenParts
            for i, part in enumerate(parts):
                lines[i].append(part.ljust(maxLenParts))
        if maxLenTotal > lineLen:
            nRow += 1
        else:
            # return '\n'.join(map(string.strip, map(string.join, lines)))
            return '\n'.join([''.join(t).strip() for t in lines])
    # unexpected long list:
    return '\n'.join(List)

hasDoubleQuotes = re.compile(r'^".*"$')
hasSingleQuotes = re.compile(r"^'.*'$")

def convertToPythonArgs(text):
    """convert to numbers and strings

    IF argument is enclosed in " " or ' ' it is kept as a string.
    """    
    text = text.strip()
    if not text:
        return None
    L = text.split(',')
    L = list(map(_convertToPythonArg, L))
    return tuple(L)

def convertToPythonArgsKwargs(text):
    """convert to numbers and strings,

    IF argument is enclosed in " " or ' ' it is kept as a string.

    also do kwargs now...
>>> convertToPythonArgsKwargs('')
((), {})
>>> convertToPythonArgsKwargs('hello')
(('hello',), {})
>>> convertToPythonArgsKwargs('width=50')
((), {'width': 50})
>>> convertToPythonArgsKwargs('"hello", width=50')
(('hello',), {'width': 50})

    

    """
    L = []
    K = {}
    text = text.strip()
    if not text:
        return tuple(L), K 
    textList = text.split(',')
    for arg in textList:
        if arg.find("=") > 0:
            kw, arg2 = [t.strip() for t in arg.split("=", 1)]
            if kw.find(" ") == -1:
                arg2 = _convertToPythonArg(arg2)
                K[kw] = arg2
            else:
                L.append(arg.strip())
        else:
            L.append(_convertToPythonArg(arg))
    return tuple(L), K

def _convertToPythonArg(t):
    t = t.strip()
    if not t:
        return ''

    # if input string is a number, return string directly
    try:
        i = int(t)
        if t == '0':
            return 0
        if t.startswith('0'):
            print(f'warning convertToPythonArg, could be int, but assume string: {t}')
            return str(t)
        return i
    except ValueError:
        pass
    try:
        f = float(t)
        if t.find(".") >= 0:
            return f
        print(f'warning convertToPythonArg, can be float, but assume string: {t}')
        return str(t)
    except ValueError:
        pass

    # now proceeding with strings:    
    if hasDoubleQuotes.match(t):
        return t[1:-1]
    if hasSingleQuotes.match(t):
        return t[1:-1]
    return t

def splitList(L, n):
    """generator function that splits a list in sublists of length n

    """
    O = []
    for l in L:
        O.append(l)
        if len(O) == n:
            yield O
            O = []
    if O:
        yield O

def makeReadable(t):
    """squeeze text for readability
    
    helper for print lines...
    """
    t = t.strip()
    t = t.replace('\n', '\\\\')
    if len(t) > 100:
        return t[:50] + ' ... ' + t[-50:]
    return t

## functions for generating alternative paths in virtual drives
## uses reAltenativePaths, defined in the top of this module
## put in utilsqh.py! used in sitegen AND in _folders.py grammar of Unimacro:
# for alternatives in virtual drive definitions:
reAltenativePaths = re.compile(r"(\([^|()]+?(\|[^|()]+?)+\))")

def generate_alternatives(s):
    """generates altenatives if (xxx|yyy) is found, otherwise just yields s
    Helper for cross_loop_alternatives
    """
    m = reAltenativePaths.match(s)
    if m:
        alternatives = s[1:-1].split("|")
        for item in alternatives:
            yield item
    else:
        yield s
        
def cross_loop_alternatives(*sequences):
    """helper function for loop_through_alternative_paths
    """
    if sequences:
        for x in generate_alternatives(sequences[0]):
            for y in cross_loop_alternatives(*sequences[1:]):
                yield (x,) + y
    else:
        yield ()

def loop_through_alternative_paths(pathdefinition):
    """can hold alternatives (a|b)
    
>>> list(loop_through_alternative_paths("(C|D):/xxxx/yyyy"))
['C:/xxxx/yyyy', 'D:/xxxx/yyyy']
>>> list(loop_through_alternative_paths("(C:|D:|E:)/Do(kum|cum)ent(s|en)"))
['C:/Dokuments', 'C:/Dokumenten', 'C:/Documents', 'C:/Documenten', 'D:/Dokuments', 'D:/Dokumenten', 'D:/Documents', 'D:/Documenten', 'E:/Dokuments', 'E:/Dokumenten', 'E:/Documents', 'E:/Documenten']

    So "(C|D):/natlink" first yields "C:/natlink" and then "D:/natlink".
    More alternatives in one item are possible, see second example.
    """
    m = reAltenativePaths.search(pathdefinition)
    if m:
        result = reAltenativePaths.split(pathdefinition)
        result = [x for x in result if x and not x.startswith("|")]
        for pathdef in cross_loop_alternatives(*result):
            yield ''.join(pathdef)
    else:
        # no alternatives, simply yield the pathdefinition:
        yield pathdefinition

def getValidPath(variablePathDefinition):
    r"""check the different alternatives of the definition
    
# >>> homefolder = getValidPath(r'(A|B|D|C|D):\Users\(Dell|Gebruiker)\.Natlink')
# >>> homefolder
# WindowsPath('C:/Users/Gebruiker/.Natlink')
# >>> str(homefolder)
# 'C:\\Users\\Gebruiker\\.Natlink'

    return the first valid path, None if not found
    """
    for p in loop_through_alternative_paths(variablePathDefinition):
        if os.path.exists(p):
            return Path(p)
    return None

def _test():
    #pylint:disable=C0415
    import doctest
    return doctest.testmod()

if __name__ == "__main__":


    def revTrue(t):
        return "y"
    def revFalse(t):
        return "n"
    def revAbort(t):
        return
    _test()



"""actions from application pythonwin
"""
#pylint:disable=R0903
import dtactions.messagefunctions as mf
from .actionbases import MessageActions
class PythonwinActions(MessageActions):
    handlesDict = {}
    def __init__(self, progInfo):
        MessageActions.__init__(self, progInfo)
        
    def getInnerHandle(self, topHndle):
        
        controls = mf.findControls(topHndle, wantedClass="Scintilla")
        print(f'Scintilla controls: {controls}')
        for c in controls:
            ln = self.getCurrentLineNumber(c)
            numberLines = self.getNumberOfLines(c)
            visible1 = mf.isVisible(c)
            print(f'control: {c}, linenumber: {ln}, nl: {numberLines}, visible: {visible1}')
                
if __name__ == '__main__':
    pass



"""actions from application frescobaldi, the editor for lilypond music note printing

now works with copy and so getting the wanted contents of text around the cursor.

"""
#pylint:disable=C0209
import time
from dtactions.unimacro.actionclasses.actionbases import AllActions
from dtactions.unimacro import unimacroutils
from dtactions.sendkeys import sendkeys

class FrescobaldiActions(AllActions):
    def __init__(self, progInfo):
        AllActions.__init__(self, progInfo)
        
    def hasSelection(self):
        """returns the text if text is selected, otherwise None
        """
        unimacroutils.saveClipboard()
        self.playString("{ctrl+c}")
        t = unimacroutils.getClipboard() or None
        unimacroutils.restoreClipboard()
        return t

    def playString(self, t):
        """send through to natlinkutils.self.playString
        """
        sendkeys(t)

    def getClipboard(self, sleep=0.05):
        """wait little longer if no result
        """
        for _ in range(5):
            time.sleep(sleep)
            clip = unimacroutils.getClipboard()
            if clip:
                return clip
        print('got nothing on clipboard')
        return ""

    def getPrevNext(self, n=1, sleep=0.05):
        """return character to the left and to the right of the cursor
        assume no selection active.
        normally return cursor in same position
        """
        unimacroutils.saveClipboard()
        self.playString("{left %s}"% n)
        self.playString("{shift+right %s}"% (n*2,))
        time.sleep(sleep)
        self.playString("{ctrl+c}")
        result = self.getClipboard(sleep=sleep)
        self.playString("{left}{right %s}"% n)
        time.sleep(sleep)
        unimacroutils.restoreClipboard()
        if not result:
            print('nothing on clipboard')
            return ""
       
        if len(result) == n*2:
            return result[:n], result[n:]
        if result == '\n':
            print('getPrevNext, assume at end of file...')
            # assume at end of file, could also be begin of file, but too rare too handle
            self.playString("{right}")
            return result, result
        print(f'getPrevNext, len not 2: {len(result)}, ({result})')
        return "", result
        
        
    def getNextLine(self, n=1):
        """get line following the cursor

        n > 1, take more lines down for a larger range
        """
        self.playString("{shift+down %s}{ctrl+c}"% n)
        unimacroutils.Wait()        
        result = unimacroutils.getClipboard()
        nup = result.count('\n')
        if nup:
            self.playString("{shift+up %s}"% nup)
        return result
        
    def getPrevLine(self, n=1):
        """get line preceding the cursor
        
        more lines possible, if n > 1
        """
        self.playString("{shift+up %s}{ctrl+c}"% n)
        unimacroutils.Wait()        
        result = unimacroutils.getClipboard()
        ndown = result.count('\n')
        if ndown:
            self.playString("{shift+down %s}"% ndown)
        return result
    
    def getWindowText(self):
        pass
    
        
                   
    
if __name__ == '__main__':
    # search all frescobaldi instances and dump the controls:
    # does not give a useful result
    from dtactions import messagefunctions as mf
    # trying to hack into frescobaldi, no luck (yet)
    tw = mf.findTopWindows(wantedText="frescobaldi")
    for tt in tw:
        dw = mf.dumpWindow(tt)
        if dw:
            print(f'top: {tt}, length dump: {len(dw)}%s'% (tt, len(dw)))
            for subwindow in dw:
                print(f'sub: {subwindow}')
                hndle = subwindow[0]
                subhndle = subwindow[-1][0]
                try:
                    numl = mf.getNumberOfLines(hndle, classname='Scintilla')
                    print(f'hndle: {hndle}, subhndle: {subhndle}, numlines: {numl}')
                except TypeError:
                    print(f'wrong hndle: {subhndle}')
        else:
            print('try topwindow')
            hndle = tt
            try:
                numl = mf.getNumberOfLines(hndle, classname='Scintilla')
                print(f'top hndle: {hndle}, numlines: {numl}')
            except TypeError:
                print(f'wrong hndle: {subhndle}')
            
            print(f'top: {tt}, mp dumpwindow')
        # pprint.pprint( mf.dumpWindow(t) )
    ## get correct window handle with "give window info" (_general grammar)
    ProgInfo = ('frescobaldi', 'naamloos', 'top', 396616)
    time.sleep(3)
    fa = FrescobaldiActions( ProgInfo )
    fa.getPrevNext()
    for j in range(1,11):
        for i in range(10):
            Result = fa.getPrevNext(j)
            if Result:
                print(f'{Result}, ({len(Result)})')
            else:
                print('no result')
            fa.playString("{left 3}")
        


"""actions from application win32pad

see http://www.gena01.com/win32pad/, the only select and say notepad like window with line numbers
"""
#pylint:disable=R0903
import pprint
import dtactions.messagefunctions as mf
from .actionbases import MessageActions


class Win32padActions(MessageActions):
    def __init__(self, progInfo):
        MessageActions.__init__(self, progInfo)
        
    def getInnerHandle(self, topHndle):
        findValues = mf.findControls(topHndle, wantedClass="RichEdit20A")
        if findValues:
            print(f'win32pad, topHndle: {topHndle}, innerHandle (take first: {findValues}')
            return findValues[0]
        # else probably a child window
        return None
    #def metaaction_gotoline(self, n):
    #    """go to a specified line directly, no calling of dialog
    #    """
        
    
if __name__ == '__main__':
    # search all win32pad instances and dump the controls:
    tw = mf.findTopWindows(wantedText="win32pad")
    for th in tw:
        print(f'top: {th}')
        pprint.pprint( mf.dumpWindow(th) )
        
        


### 
### Module Keys
### 

from dtactions.vocola_sendkeys import ExtendedSendDragonKeys
from dtactions.vocola_sendkeys import SendInput



# Vocola procedure: Keys.SendInput
def send_input(specification):
    SendInput.send_input(
        ExtendedSendDragonKeys.senddragonkeys_to_events(specification))



"""actions from application uedit32, UltraEdit

(from IDM Computer Solutions, www.ultraedit.com)
"""
#pylint:disable=R0903
import win32gui
import dtactions.messagefunctions as mf
from .actionbases import MessageActions

class Uedit32Actions(MessageActions):
    def __init__(self, progInfo):
        MessageActions.__init__(self, progInfo)
        
    def getInnerHandle(self, topHndle):
        """procedure to get inner handle for EditControl,
        
        via a control that holds the filename, which is extracted from the window title
        """
        title = win32gui.GetWindowText(topHndle)
        #print 'title: %s'% title
        tabName = title.split(']')[0]
        tabName = tabName.split('[')[-1]
        self.prevTabName = tabName
        #print 'tabName: %s'% tabName
        controls = mf.findControls(topHndle, wantedText = tabName)
        if not controls:
            #print 'uedit32Actions, no control with title: "%s" found'% tabName
            return None
        if len(controls) > 1:
            print(f'uedit32Actions, strange, found more controls, matching the title: "{tabName}", take first')
        c = controls[0]
        innerControl = mf.findControls(c, wantedClass="EditControl")[0]
        return innerControl
    
if __name__ == '__main__':

    # test, fill in handle of ultraedit (uedit32) (with "give window info" from grammar _general)
    focus =199852
    print(f'focus; {focus}')
    ProgInfo = ('uedit32', 'babbababa', 'top', focus)
    ue = Uedit32Actions( ProgInfo )



#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
#pylint:disable=E0611, E0401, R0911, R0912, W0702, C0116
#pylint:disable=C0209
"""extenvvars.py

Keep track of environment variables, including added "fake" variables like DROPBOX

 Quintijn Hoogenboom, January 2008/September 2015/November 2022 (python3)/March 24


"""
import os
from os.path import normpath, isfile, isdir, join
import copy
import re
import ctypes
from pathlib import Path
from win32com.shell import shellcon

import platformdirs
try:
    import natlink
    natlinkAvailable = True
except ImportError:
    natlinkAvailable = False
    
if natlinkAvailable:
    if not natlink.isNatSpeakRunning():
        natlinkAvailable = False
        
if natlinkAvailable:
    from natlinkcore import natlinkstatus
    status = natlinkstatus.NatlinkStatus()
    status_dict_full = copy.copy(status.getNatlinkStatusDict())
    
    status_dict = {key.upper():value for (key, value) in status_dict_full.items() if os.path.isdir(value)}
    del status_dict_full
else:
    status = None
    status_dict = None
# for extended environment variables:
reEnv = re.compile('(%[A-Z_]+%)', re.I)

class ExtEnvVars:
    """gives and "remembers" environment variables, with extensions
    
    several sorts of env variables are handled:
    
    Environment variables, like %xxxx%:
    - os.environ
    - all from natlinkstatus.py (if available) (like %natlink% or %unimacro%)


    Libraries, that are returned from a getFolderFromActiveWindow (_folders grammar of unimacro):
        - several from "Libraries" (like Music, Documents), via platformdirs
        - others from "CSIDL_variables" directly
    """
    
    def __init__(self):
        self.recentEnv = {}
        self.homeDir = self.getHome()        # starting at least reventEnv dict.

    def addToRecentEnv(self, name, value):
        """to be filled for NATLINK variables from natlinkstatus and other env variables, sort of caching mechanism
        
        empty values can be cached...
        """
        if not value:
            self.recentEnv[name] = None
            return
        value = str(value)
        if not isdir(value):
            self.recentEnv[name] = None
            raise ValueError('extenvvars, addToRecentEnv: not a valid path for for name: "%s": "%s"'% (name, value))
        self.recentEnv[name] = value

    def deleteFromRecentEnv(self, name):
        """to possibly delete from recentEnv, from natlinkstatus
        """
        if name in self.recentEnv:
            del self.recentEnv[name]

    def getFolderFromLibraryName(self, name):
        """from windows library names extract the real folder
        
        the CabinetWClass and #32770 controls return "Documents", "Dropbox", "Desktop" etc.
        try to resolve these throug the extended env variables,
        - with platformdirs
        - with more special trick
        """
        if name in self.recentEnv:
            return self.recentEnv[name]
        
        if name == "~":
            home_dir = Path().home()
            self.addToRecentEnv("~", home_dir)
            return self.recentEnv["~"]
        
        
        ## these are NOT cached:
        if name.startswith("Docum"):  # Documents in Dutch and English
            return platformdirs.windows.get_win_folder("CSIDL_PERSONAL")
        if name in ["Muziek", "Music"]:
            return platformdirs.windows.get_win_folder("CSIDL_MYMUSIC")
        if name in ["Pictures", "Afbeeldingen"]:
            return platformdirs.windows.get_win_folder("CSIDL_MYPICTURES")
        if name in ["Videos", "Video's"]:
            return platformdirs.windows.get_win_folder("CSIDL_MYVIDEO")
        if name in ['Desktop', "Bureaublad"]:
            return platformdirs.windows.get_win_folder("CSIDL_DESKTOPDIRECTORY")
        if name in ['Download', 'Downloads']:
            return platformdirs.windows.get_win_folder("CSIDL_DOWNLOADS")
        if name in ['APPDATA']:
            return platformdirs.windows.get_win_folder("CSIDL_APPDATA")
        if name in ['COMMON_APPDATA']:
            return platformdirs.windows.get_win_folder("CSIDL_COMMON_APPDATA")
        if name in ['LOCAL_APPDATA']:
            return platformdirs.windows.get_win_folder("CSIDL_LOCAL_APPDATA")

        # # General case, try via shellcon! TODO: QH
        # try:
        #     CSIDL_variable =  'CSIDL_%s'% name
        #     shellnumber = getattr(shellcon,CSIDL_variable, -1)
        #     print(f'getFolderFromLibraryName, shellnumber of "{CSIDL_variable}": {shellnumber}')
        # except:
        #     print('getExtendedEnv, cannot find CSIDL_variable for: "%s"'% name)
        #     return ''
        # if shellnumber < 0:
        #     # on some systems have SYSTEMROOT instead of SYSTEM:
        #     print('getExtendedEnv, cannot find CSIDL_variable for (returns -1): "%s"'% name)
        # try:
        #     csidl_const = shellnumber
        #     # copied from platformdirs/windows.py:
        #     buf = ctypes.create_unicode_buffer(1024)
        #     windll = getattr(ctypes, "windll")  # noqa: B009 # using getattr to avoid false positive with mypy type checker
        #     windll.shell32.SHGetFolderPathW(None, csidl_const, None, 0, buf)
        #     result = buf.value
        #     # # result = shell.SHGetFolderPath (0, shellnumber, 0, 0)
        #     # result = ctypes.windll.shell32.SHGetFolderPathW(0, shellnumber, 0, 0)
        #     print(f'result from SHGetFolderPathW: {result}')
        # except:
        #     print('getFolderFromLibraryName, cannot path for CSIDL: "%s"'% name)
        #     return ''

        
        ## extra cases:
        # if name in ['Quick access', 'Snelle toegang']:
        #     templatesfolder =self. getExtendedEnv('TEMPLATES')
        #     if isdir(templatesfolder):
        #         QuickAccess = normpath(join(templatesfolder, "..", "Libraries"))
        #         if isdir(QuickAccess):
        #             return QuickAccess
        if name.upper() == 'DROPBOX':
            return self.getDropboxFolder()
        if name.upper() in ['ONEDRIVE']:
            return self.getExtendedEnv("OneDrive")
        
        if name in ["This PC", "Deze pc"]:
            return "\\"
        print('getFolderFromLibraryName, cannot find folder for Library name: %s'% name)
        return ""

    def getHome(self):
        """get the home directory, with pathlib.Path().home()
        
        If env variable HOME is there, and points to another directory, give a warning. 
        
        result is cached in recentEnv dict.
        
        """
        if "~" in self.recentEnv:
            return self.recentEnv["~"]

        Home = str(Path().home())
        if Home and isdir(Home):
            self.addToRecentEnv("~", Home)   # cache as a str
        HomeFromEnv = os.environ.get("HOME", "")
        if HomeFromEnv and isdir(HomeFromEnv):
            self.addToRecentEnv("HOME", HomeFromEnv)
            if HomeFromEnv != Home:
                print(f'Warning: "~" and "HOME" point to different directories\n\t"~": "{Home}"\n\t"HOME": "{HomeFromEnv}"')
        else:
            self.addToRecentEnv("HOME", Home)
        return self.recentEnv["~"]
            

    def getDropboxFolder(self):
        """get the dropbox folder, or the subfolder which is specified.
        
        searched is in home directory of the user...
        
        raises OSError if more folders are found (should not happen, I think)
        if containsFolder is not passed, the dropbox main folder is returned
        if containsFolder is passed, this folder is returned if it is found in the dropbox folder
        
        otherwise None is returned.
        """
        if "DROPBOX" in self.recentEnv:
            return self.recentEnv["DROPBOX"]

        dirsToTry = [ "%DROPBOX%", "~", 'C:\\']    #### for s in os.listdir(root) if isdir(join(root,s)) ]
        for root in dirsToTry:
            result = self.getDropboxFolder2(root)
            if result:
                self.addToRecentEnv("DROPBOX", result)
                return self.recentEnv["DROPBOX"]
        return False 
            
    def getDropboxFolder2(self, root):
        """helper function
        """
        
        if root.startswith("%"):
            root = root.strip("% ").upper()
            root = os.environ.get(root, "")
        
        if root and not isdir(root):
            return False
        
        if root.lower().endswith("dropbox"):
            return root
        next_dir = join(root, "dropbox")
        if isdir(next_dir):
            return next_dir
        return False
    
    def matchesStart(self, listOfDirs, checkDir, caseSensitive):
        """return result from list if checkDir matches, mostly case insensitive
        """
        if not caseSensitive:
            checkDir = checkDir.lower()
        for l in listOfDirs:
            if not caseSensitive:
                ll = l.lower()
            else:
                ll = l
            if ll.startswith(checkDir):
                return l
        return False   
        
            

    def getExtendedEnv(self, var, noCache=False, displayMessage=True):
        r"""get from os.environ or windows CSLID
    
        HOME is os.environ['HOME'] or CSLID_PERSONAL
        ~ is HOME (this is doubtful, str(pathlib.Path.home()) gives 'C:\Users\User', also when HOME points to another directory
    
        DROPBOX added via self.getDropboxFolder is this module (QH, December 1, 2018)
    
        Also the directories that are configured inside Natlink can be retrieved, via natlinkstatus.py
        Natlink_SettingsDir, DragonflyUserDirectory, VocolaDirectory, AhkUserDir:
        
        Note: these settings are case sensitive! You can leave out Dir or Directory.
    
        """
        var = var.strip("% ")
    
        if var in  self.recentEnv:
            return self.recentEnv[var]
    
        # Note the Natlink variables must conform to "Unimacro" etc. Best don't append "Dir" or "Directory":
        # It is tried to append ("Dir" or "Directory" (or nothing) to it in the call
        # Don't capitalise these variables!!)
        result = self.getDirectoryFromNatlinkstatus(var)
        if result:
            self.recentEnv[var] = result
            return result
      
        
        
        if var == "~":
            var = 'HOME'
        
        if var in os.environ:
            result = os.environ[var]
            self.addToRecentEnv(var, result)
            return self.recentEnv[var]
    
        if var == 'DROPBOX':
            result = self.getDropboxFolder()
            if result:
                self.addToRecentEnv(var, result)
                return result
            raise ValueError('getExtendedEnv, cannot find path for "DROPBOX"')
    
        if var == 'NOTEPAD':
            windowsDir =self. getExtendedEnv("WINDOWS")
            notepadPath = join(windowsDir, 'notepad.exe')
            if isfile(notepadPath):
                return notepadPath
            raise ValueError('getExtendedEnv, cannot find path for "NOTEPAD"')
    
        # try to get from CSIDL system call:
        try:
            CSIDL_variable =  'CSIDL_%s'% var
            shellnumber = getattr(shellcon,CSIDL_variable, -1)
        except:
            print('getExtendedEnv, cannot find in os.environ or CSIDL: "%s"'% var)
            return ''
        if shellnumber < 0:
            # on some systems have SYSTEMROOT instead of SYSTEM:
            if var == 'SYSTEM':
                return self.getExtendedEnv('SYSTEMROOT')
            return ''
        try:
            csidl_const = shellnumber
            # copied from platformdirs/windows.py:
            buf = ctypes.create_unicode_buffer(1024)
            windll = getattr(ctypes, "windll")  # noqa: B009 # using getattr to avoid false positive with mypy type checker
            windll.shell32.SHGetFolderPathW(None, csidl_const, None, 0, buf)
            result = buf.value
            # # result = shell.SHGetFolderPath (0, shellnumber, 0, 0)
            # result = ctypes.windll.shell32.SHGetFolderPathW(0, shellnumber, 0, 0)
            print(f'result from SHGetFolderPathW: {result}')
        except:
            if displayMessage:
                print('getExtendedEnv, cannot find in os.environ or CSIDL: "%s"'% var)
            return ''
        if noCache:
            return result
        self.addToRecentEnv(var, result)
        return self.recentEnv[var]

    def getDirectoryFromNatlinkstatus(self, envvar):
        """see if directory can can be retrieved from envvar
        """
        # if natlink not available:
        if not natlinkAvailable:
            # print(f'natlink not available for get "{envvar}"')
            return None
    
        # try if function in natlinkstatus:
        if not status:
            return None
        for extra in ('', 'Directory', 'Dir'):
            var2 = envvar + extra
            if var2 in status.__dict__:
                funcName = f'get{var2}'
                func = getattr(status, funcName)
                result = func()
                if result:
                    return result
        return None
    
    
    
    def clearRecentEnv(self):
        """for testing, clears above global dictionary
        """
        self.recentEnv.clear()
    
    def getAllFolderEnvironmentVariables(self, displayMessage=False):
        """return, as a dict, all the os.environ AND all CSLID variables that result into a folder
        
        Now also implemented:  Also include NATLINK, UNIMACRO, VOICECODE, DRAGONFLY, VOCOLAUSERDIR, UNIMACROUSERDIR
        This is done by calling from natlinkstatus, see there and example in natlinkmain.
    
        Optionally put them in recentEnv, if you specify fillRecentEnv to 1 (True)
    
        """
        D = {}
    
        for k in dir(shellcon):
            if k.startswith("CSIDL_"):
                kStripped = k[6:]
                try:
                    v =self.getExtendedEnv(kStripped, noCache=True, displayMessage=displayMessage)    ## displayMessage=False)
                except ValueError:
                    continue
                if len(str(v)) > 2 and isdir(v):
                    D[kStripped] = v
        # os.environ overrules CSIDL:
        for k, v in os.environ.items():
            if isdir(v):
                v = normpath(v)
                if k in D and D[k] != v:
                    print('warning, CSIDL also exists for key: %s, take os.environ value: %s'% (k, v))
                D[k] = v
        return D
    def substituteEnvVariableAtStart(self, filepath, envDict=None): 
        r"""try to substitute back one of the (preused) environment variables back
    
        into the start of a filename
    
        if ~ (HOME) is D:\My documents,
        the path "D:\My documents\folder\file.txt" should return "~\folder\file.txt"
    
        pass in a dict of possible environment variables, which can be taken from recent calls, or
        from  envDict = getAllFolderEnvironmentVariables().
    
        Alternatively you can call getAllFolderEnvironmentVariables once, and use the recentEnv
        of this module! getAllFolderEnvironmentVariables(fillRecentEnv)
    
        If you do not pass such a dict, recentEnv is taken, but this recentEnv holds only what has been
        asked for in the session, so no complete list!
    
        """
        if envDict is None:
            envDict = self.recentEnv
        Keys = list(envDict.keys())
        # sort, longest result first, shortest keyname second:
        decorated = [(-len(envDict[k]), len(k), k) for k in Keys]
        decorated.sort()
        Keys = [k for (dummy1,dummy2, k) in decorated]
        for k in Keys:
            val = envDict[k]
            if filepath.lower().startswith(val.lower()):
                if k in ("HOME", "PERSONAL"):
                    k = "~"
                else:
                    k = "%" + k + "%"
                filepart = filepath[len(val):]
                filepart = filepart.strip('/\\ ')
                return join(k, filepart)
        # no hit, return original:
        return filepath
           
    def expandEnvVariableAtStart(self, filepath, envDict=None): 
        """try to substitute environment variable into a path name
    
        """
        filepath = filepath.strip()
    
        if filepath.startswith('~'):
            folderpart =self. getExtendedEnv('~', envDict)
            filepart = filepath[1:]
            filepart = filepart.strip('/\\ ')
            return normpath(join(folderpart, filepart))
        if reEnv.match(filepath):
            envVar = reEnv.match(filepath).group(1)
            # get the envVar...
            try:
                folderpart =self. getExtendedEnv(envVar, envDict)
            except ValueError:
                print('invalid (extended) environment variable: %s'% envVar)
            else:
                # OK, found:
                filepart = filepath[len(envVar)+1:]
                filepart = filepart.strip('/\\ ')
                return normpath(join(folderpart, filepart))
        # no match
        return filepath
        
    def expandEnvVariables(self, filepath, envDict=None): 
        """try to substitute environment variable into a path name,
    
        ~ only at the start,
    
        %XXX% can be anywhere in the string.
    
        """
        filepath = filepath.strip()
        
        if filepath.startswith('~'):
            folderpart =self. getExtendedEnv('~', envDict)
            filepart = filepath[1:]
            filepart = filepart.strip('/\\ ')
            filepath = normpath(join(folderpart, filepart))
        
        if reEnv.search(filepath):
            List = reEnv.split(filepath)
            #print 'parts: %s'% List
            List2 = []
            for part in List:
                if not part:
                    continue
                if part == "~" or (part.startswith("%") and part.endswith("%")):
                    try:
                        folderpart = self.getExtendedEnv(part, envDict)
                    except ValueError:
                        folderpart = part
                    List2.append(folderpart)
                else:
                    List2.append(part)
            List2 = [str(item) for item in List2]
            filepath = ''.join(List2)
            return normpath(filepath)
        # no match
        return filepath
    
    
    
    def printAllEnvVariables(self):
        for k in sorted(self.recentEnv.keys()):
            print("%s\t%s"% (k, self.recentEnv[k]))
    
if __name__ == "__main__":
    env = ExtEnvVars()
    Vars = env.getAllFolderEnvironmentVariables()
    print('recentEnv: ')
    print(env.recentEnv)
    print('='*80)
    

    print('testing       expandEnvVariableAtStart')
    print('also see expandEnvVar in natlinkstatus!! and test_extenvvars.py in dtactions/test')
    for p in ("~", "%home%", "D:\\natlink\\unimacro", "~/unimacroqh",
              "%HOME%/personal",
              "%WINDOWS%\\folder\\strange testfolder"):
        expanded = env.expandEnvVariableAtStart(p)
        print('expandEnvVariablesAtStart: %s: %s'% (p, expanded))
        
    print('testing       expandEnvVariables')  
    p = "%UNIMACROUSER%"
    expanded = env.expandEnvVariables(p)


    for p in ("%NATLINK%\\unimacro", "%DROPBOX%/QuintijnHerold/jachthutten", "D:\\%username%",  "%UNIMACROUSER%", "%HOME%/personal", "%HOME%", "%personal%"
              "%WINDOWS%\\folder\\strange testfolder"):
        expanded = env.expandEnvVariables(p)
        print('expandEnvVariables: %s: %s'% (p, expanded))





### 
### Interface to Microsoft Windows SendInput function using ctypes
### (requires Python 2.5+)
### 
###     This code allows sending an arbitrary sequence of raw input
### events via the SendInput call.  Determining the correct sequence
### of raw input events to produce a desired result based on the
### current keyboard layout and key state is the responsibility of the
### caller.
### 
### 
### Author:  Mark Lillibridge
### Version: 0.8
### 
#pylint:disable = W0622, R0903, C0321, 

from ctypes import *
import win32con

## 
## SendInput function:
## 
##   Synthesizes keystrokes, mouse motions, and button clicks.
## 
##       Events is a list of objects that support a to_input() method
##   that returns an Input.  Events can be created directly from the
##   raw data structures ({Mouse,Keyboard,Hardware}Input) or using
##   event-creation convenience functions.
## 
##       Can be blocked by other threads (e.g., BlockInput) or UIPI
##   (applications are not permitted to inject input into applications
##   at a higher integrity level).  The latter causes silent failure.
## 
##       These events are not interspersed with other keyboard or mouse
##   input events inserted either by the user (with the keyboard or
##   mouse) or by calls to keybd_event, mouse_event, or other calls to
##   SendInput.
##
##       This function does not reset the keyboard's current state. Any
##   keys that are already pressed when the function is called might
##   interfere with the events that this function generates. To avoid
##   this problem, check the keyboard's state with the GetAsyncKeyState
##   function and correct as necessary.
## 

def send_input(events):
    inputs = [e.to_input() for e in events]
    input = (Input * len(events))(*inputs)
    inserted = windll.user32.SendInput(len(input), byref(input), sizeof(Input))
    if inserted != len(events):
        raise ValueError("windll.user32.SendInput: ???")  ## + FormatMessage())


## 
## The raw INPUT data structure used to pass events to SendInput.
## 

DWORD     = c_ulong          # 32 bits
LONG      = c_long           # 32 bits
ULONG     = c_ulong          # 32 bits 
ULONG_PTR = POINTER(ULONG)
WORD      = c_ushort         # 16 bits

class MouseInput(Structure):
    _fields_ = [('dx',          LONG),
                ('dy',          LONG),
                ('mouseData',   DWORD),
                ('dwFlags',     DWORD),
                ('time',        DWORD),
                ('dwExtraInfo', ULONG_PTR)]
    def to_input(self):
        return Input(win32con.INPUT_MOUSE, _EventUnion(mi=self))

class KeyboardInput(Structure):
    _fields_ = [('wVk',         WORD),
                ('wScan',       WORD),
                ('dwFlags',     DWORD),
                ('time',        DWORD),
                ('dwExtraInfo', ULONG_PTR)]
    def to_input(self):
        return Input(win32con.INPUT_KEYBOARD, _EventUnion(ki=self))

class HardwareInput(Structure):
    _fields_ = [('uMsg',    DWORD),
                ('wParamL', WORD),
                ('wParamH', WORD)]
    def to_input(self):
        return Input(win32con.INPUT_HARDWARE, _EventUnion(hi=self))

class _EventUnion(Union):
    _fields_ = [('mi', MouseInput),
                ('ki', KeyboardInput),
                ('hi', HardwareInput)]

class Input(Structure):
    _fields_ = [('type', DWORD),
                ('Union', _EventUnion)]
    def to_input(self):
        return self



### 
### Keyboard events:
### 

## 
## All defined virtual-key codes; from:
## 
##   http://msdn.microsoft.com/en-us/library/windows/desktop/dd375731%28v=vs.85%29.aspx
## 

# Modifiers:

VK_SHIFT               = 0x10  # SHIFT key
VK_CONTROL             = 0x11  # CTRL key
VK_MENU                = 0x12  # ALT key

VK_LSHIFT              = 0xA0  # Left  SHIFT key
VK_RSHIFT              = 0xA1  # Right SHIFT key
VK_LCONTROL            = 0xA2  # Left  CONTROL key
VK_RCONTROL            = 0xA3  # Right CONTROL key
VK_LMENU               = 0xA4  # Left  MENU key
VK_RMENU               = 0xA5  # Right MENU key

VK_LWIN                = 0x5B  # Left  Windows key (Natural keyboard)
VK_RWIN                = 0x5C  # Right Windows key (Natural keyboard)
VK_APPS                = 0x5D  # Applications key (Natural keyboard)

# Toggle keyboard-state keys:

VK_CAPITAL             = 0x14  # CAPS LOCK key
VK_NUMLOCK             = 0x90  # NUM LOCK key
VK_SCROLL              = 0x91  # SCROLL LOCK key

# Mouse buttons (only used to check button state using
# Get[Async]KeyState as far as I know):

VK_LBUTTON             = 0x01  # Left mouse button
VK_RBUTTON             = 0x02  # Right mouse button
VK_MBUTTON             = 0x04  # Middle mouse button (three-button mouse)
VK_XBUTTON1            = 0x05  # X1 mouse button
VK_XBUTTON2            = 0x06  # X2 mouse button

# Alphanumeric keys:

VK_0                   = 0x30  # 0 key
VK_1                   = 0x31  # 1 key
VK_2                   = 0x32  # 2 key
VK_3                   = 0x33  # 3 key
VK_4                   = 0x34  # 4 key
VK_5                   = 0x35  # 5 key
VK_6                   = 0x36  # 6 key
VK_7                   = 0x37  # 7 key
VK_8                   = 0x38  # 8 key
VK_9                   = 0x39  # 9 key

VK_A                   = 0x41  # A key
VK_B                   = 0x42  # B key
VK_C                   = 0x43  # C key
VK_D                   = 0x44  # D key
VK_E                   = 0x45  # E key
VK_F                   = 0x46  # F key
VK_G                   = 0x47  # G key
VK_H                   = 0x48  # H key
VK_I                   = 0x49  # I key
VK_J                   = 0x4A  # J key
VK_K                   = 0x4B  # K key
VK_L                   = 0x4C  # L key
VK_M                   = 0x4D  # M key
VK_N                   = 0x4E  # N key
VK_O                   = 0x4F  # O key
VK_P                   = 0x50  # P key
VK_Q                   = 0x51  # Q key
VK_R                   = 0x52  # R key
VK_S                   = 0x53  # S key
VK_T                   = 0x54  # T key
VK_U                   = 0x55  # U key
VK_V                   = 0x56  # V key
VK_W                   = 0x57  # W key
VK_X                   = 0x58  # X key
VK_Y                   = 0x59  # Y key
VK_Z                   = 0x5A  # Z key

# Function keys:

VK_F1                  = 0x70  # F1 key
VK_F2                  = 0x71  # F2 key
VK_F3                  = 0x72  # F3 key
VK_F4                  = 0x73  # F4 key
VK_F5                  = 0x74  # F5 key
VK_F6                  = 0x75  # F6 key
VK_F7                  = 0x76  # F7 key
VK_F8                  = 0x77  # F8 key
VK_F9                  = 0x78  # F9 key
VK_F10                 = 0x79  # F10 key
VK_F11                 = 0x7A  # F11 key
VK_F12                 = 0x7B  # F12 key
VK_F13                 = 0x7C  # F13 key
VK_F14                 = 0x7D  # F14 key
VK_F15                 = 0x7E  # F15 key
VK_F16                 = 0x7F  # F16 key
VK_F17                 = 0x80  # F17 key
VK_F18                 = 0x81  # F18 key
VK_F19                 = 0x82  # F19 key
VK_F20                 = 0x83  # F20 key
VK_F21                 = 0x84  # F21 key
VK_F22                 = 0x85  # F22 key
VK_F23                 = 0x86  # F23 key
VK_F24                 = 0x87  # F24 key

# Remaining codes sorted by name:

VK_ACCEPT              = 0x1E  # IME accept
VK_ADD                 = 0x6B  # Add key
VK_ATTN                = 0xF6  # Attn key
VK_BACK                = 0x08  # BACKSPACE key
VK_BROWSER_BACK        = 0xA6  # Browser Back key
VK_BROWSER_FAVORITES   = 0xAB  # Browser Favorites key
VK_BROWSER_FORWARD     = 0xA7  # Browser Forward key
VK_BROWSER_HOME        = 0xAC  # Browser Start and Home key
VK_BROWSER_REFRESH     = 0xA8  # Browser Refresh key
VK_BROWSER_SEARCH      = 0xAA  # Browser Search key
VK_BROWSER_STOP        = 0xA9  # Browser Stop key
VK_CANCEL              = 0x03  # Control-break processing
VK_CLEAR               = 0x0C  # CLEAR key (non-numlock version of numpad 5, a " ")
VK_CONVERT             = 0x1C  # IME convert
VK_CRSEL               = 0xF7  # CrSel key
VK_DECIMAL             = 0x6E  # Decimal key
VK_DELETE              = 0x2E  # DEL key
VK_DIVIDE              = 0x6F  # Divide key
VK_DOWN                = 0x28  # DOWN ARROW key
VK_END                 = 0x23  # END key
VK_EREOF               = 0xF9  # Erase EOF key
VK_ESCAPE              = 0x1B  # ESC key
VK_EXECUTE             = 0x2B  # EXECUTE key
VK_EXSEL               = 0xF8  # ExSel key
VK_FINAL               = 0x18  # IME final mode
VK_HANGUEL             = 0x15  # IME Hanguel mode (maintained for compatibility; use VK_HANGUL)
VK_HANGUL              = 0x15  # IME Hangul mode
VK_HANJA               = 0x19  # IME Hanja mode
VK_HELP                = 0x2F  # HELP key
VK_HOME                = 0x24  # HOME key
VK_INSERT              = 0x2D  # INS key
VK_JUNJA               = 0x17  # IME Junja mode
VK_KANA                = 0x15  # IME Kana mode
VK_KANJI               = 0x19  # IME Kanji mode
VK_LAUNCH_APP1         = 0xB6  # Start Application 1 key
VK_LAUNCH_APP2         = 0xB7  # Start Application 2 key
VK_LAUNCH_MAIL         = 0xB4  # Start Mail key
VK_LAUNCH_MEDIA_SELECT = 0xB5  # Select Media key
VK_LEFT                = 0x25  # LEFT ARROW key
VK_MEDIA_NEXT_TRACK    = 0xB0  # Next Track key
VK_MEDIA_PLAY_PAUSE    = 0xB3  # Play/Pause Media key
VK_MEDIA_PREV_TRACK    = 0xB1  # Previous Track key
VK_MEDIA_STOP          = 0xB2  # Stop Media key
VK_MODECHANGE          = 0x1F  # IME mode change request
VK_MULTIPLY            = 0x6A  # Multiply key
VK_NEXT                = 0x22  # PAGE DOWN key
VK_NONAME              = 0xFC  # Reserved
VK_NONCONVERT          = 0x1D  # IME nonconvert
VK_NUMPAD0             = 0x60  # Numeric keypad 0 key
VK_NUMPAD1             = 0x61  # Numeric keypad 1 key
VK_NUMPAD2             = 0x62  # Numeric keypad 2 key
VK_NUMPAD3             = 0x63  # Numeric keypad 3 key
VK_NUMPAD4             = 0x64  # Numeric keypad 4 key
VK_NUMPAD5             = 0x65  # Numeric keypad 5 key
VK_NUMPAD6             = 0x66  # Numeric keypad 6 key
VK_NUMPAD7             = 0x67  # Numeric keypad 7 key
VK_NUMPAD8             = 0x68  # Numeric keypad 8 key
VK_NUMPAD9             = 0x69  # Numeric keypad 9 key
VK_OEM_102             = 0xE2  # Either the angle bracket key or the backslash key on the RT 102-key keyboard
VK_OEM_1               = 0xBA  # Used for miscellaneous characters; it can vary by keyboard.  For the US standard keyboard, the ';:' key
VK_OEM_2               = 0xBF  # Used for miscellaneous characters; it can vary by keyboard.  For the US standard keyboard, the '/?' key
VK_OEM_3               = 0xC0  # Used for miscellaneous characters; it can vary by keyboard.  For the US standard keyboard, the '`~' key
VK_OEM_4               = 0xDB  # Used for miscellaneous characters; it can vary by keyboard.  For the US standard keyboard, the '[{' key
VK_OEM_5               = 0xDC  # Used for miscellaneous characters; it can vary by keyboard.  For the US standard keyboard, the '\|' key
VK_OEM_6               = 0xDD  # Used for miscellaneous characters; it can vary by keyboard.  For the US standard keyboard, the ']}' key
VK_OEM_7               = 0xDE  # Used for miscellaneous characters; it can vary by keyboard.  For the US standard keyboard, the 'single-quote/double-quote' key
VK_OEM_8               = 0xDF  # Used for miscellaneous characters; it can vary by keyboard.
VK_OEM_CLEAR           = 0xFE  # Clear key
VK_OEM_COMMA           = 0xBC  # For any country/region, the ',' key
VK_OEM_MINUS           = 0xBD  # For any country/region, the '-' key
VK_OEM_PERIOD          = 0xBE  # For any country/region, the '.' key
VK_OEM_PLUS            = 0xBB  # For any country/region, the '+' key
VK_PA1                 = 0xFD  # PA1 key
VK_PACKET              = 0xE7  # Used to pass Unicode characters as if they were keystrokes.  The VK_PACKET key is the low word of a 32-bit Virtual Key value used for non-keyboard input methods.  For more information, see Remark in KEYBDINPUT, SendInput, WM_KEYDOWN, and WM_KEYUP.
VK_PAUSE               = 0x13  # PAUSE key
VK_PLAY                = 0xFA  # Play key
VK_PRINT               = 0x2A  # PRINT key
VK_PRIOR               = 0x21  # PAGE UP key
VK_PROCESSKEY          = 0xE5  # IME PROCESS key
VK_RETURN              = 0x0D  # ENTER key
VK_RIGHT               = 0x27  # RIGHT ARROW key
VK_SELECT              = 0x29  # SELECT key
VK_SEPARATOR           = 0x6C  # Separator key
VK_SLEEP               = 0x5F  # Computer Sleep key
VK_SNAPSHOT            = 0x2C  # PRINT SCREEN key
VK_SPACE               = 0x20  # SPACEBAR
VK_SUBTRACT            = 0x6D  # Subtract key
VK_TAB                 = 0x09  # TAB key
VK_UP                  = 0x26  # UP ARROW key
VK_VOLUME_DOWN         = 0xAE  # Volume Down key
VK_VOLUME_MUTE         = 0xAD  # Volume Mute key
VK_VOLUME_UP           = 0xAF  # Volume Up key
VK_ZOOM                = 0xFB  # Zoom key


## 
## Generalized (virtual) key codes:
## 
##     These are a new generalization of virtual key codes that allow
## distinguishing between the keys that have both a numpad and
## non-numpad version using a single code.  The relevant keys and
## their generalized key codes are:
## 
##                    non-numpad:      numpad:
##   home      key     VK_HOME      GK_NUM_HOME
##   end       key     VK_END       GK_NUM_END
##   left      key     VK_LEFT      GK_NUM_LEFT
##   right     key     VK_RIGHT     GK_NUM_RIGHT
##   up        key     VK_UP        GK_NUM_UP
##   down      key     VK_DOWN      GK_NUM_DOWN
##   page up   key     VK_PRIOR     GK_NUM_PRIOR
##   page down key     VK_NEXT      GK_NUM_NEXT
##   insert    key     VK_INSERT    GK_NUM_INSERT
##   delete    key     VK_DELETE    GK_NUM_DELETE
##   return    key     VK_RETURN    GK_NUM_RETURN
## 
## All other keys' generalized key codes are the same as their virtual
## key codes.
## 
##     Note that some software does not handle the numpad versions
## correctly in some cases due to bugs (e.g., shift left typed via the
## numpad fails to select) so it is best to stick with the non-numpad
## versions unless the application intentionally assigns them
## different meanings.
## 

USE_NUMPAD = 0x1000

GK_NUM_HOME   = VK_HOME   + USE_NUMPAD
GK_NUM_END    = VK_END    + USE_NUMPAD
GK_NUM_LEFT   = VK_LEFT   + USE_NUMPAD
GK_NUM_RIGHT  = VK_RIGHT  + USE_NUMPAD
GK_NUM_UP     = VK_UP     + USE_NUMPAD
GK_NUM_DOWN   = VK_DOWN   + USE_NUMPAD
GK_NUM_PRIOR  = VK_PRIOR  + USE_NUMPAD
GK_NUM_NEXT   = VK_NEXT   + USE_NUMPAD
GK_NUM_INSERT = VK_INSERT + USE_NUMPAD
GK_NUM_DELETE = VK_DELETE + USE_NUMPAD
GK_NUM_RETURN = VK_RETURN + USE_NUMPAD


# 
# Which virtual key codes require the extended key bit on is basically
# historical: if it wasn't on the original IBM keyboard then I think
# it needs the extended keyboard bit on.
# 
Extended_bit = {}  # Default: extended bit is not allowed
for key in [VK_HOME, VK_END, VK_NEXT, VK_PRIOR, VK_INSERT, VK_DELETE, 
            VK_LEFT, VK_RIGHT, VK_UP, VK_DOWN]:
    # Case 1: extended bit is optional (two versions exist) but on for
    # the non-numpad version.
    Extended_bit[key] = 1

# Case 2: extended bit is optional (two versions exist) but off for
# the non-numpad version.
Extended_bit[VK_RETURN] = 2

for key in [VK_RCONTROL, VK_RMENU, VK_LWIN, VK_RWIN, VK_APPS, VK_SNAPSHOT, 
            VK_CANCEL, VK_NUMLOCK, VK_DIVIDE,  
            VK_VOLUME_DOWN, VK_VOLUME_MUTE, VK_VOLUME_UP, 
            VK_MEDIA_NEXT_TRACK, VK_MEDIA_PLAY_PAUSE, VK_MEDIA_PREV_TRACK, 
            VK_MEDIA_STOP, 
            VK_BROWSER_BACK, VK_BROWSER_FAVORITES, VK_BROWSER_FORWARD, 
            VK_BROWSER_HOME, VK_BROWSER_REFRESH, VK_BROWSER_SEARCH, 
            VK_BROWSER_STOP, 
            VK_LAUNCH_APP1, VK_LAUNCH_APP2, VK_LAUNCH_MAIL, 
            VK_LAUNCH_MEDIA_SELECT]:
    Extended_bit[key] = 3   # Case 3: extended bit is required


def unpack_generalized_key_code(generalized_key_code):
    virtual_key_code = generalized_key_code
    numpad_requested = False
    if generalized_key_code >= USE_NUMPAD:
        virtual_key_code -= USE_NUMPAD
        numpad_requested = True
    
    try:
        case = Extended_bit[virtual_key_code]
        if   case == 1: extended_bit = not numpad_requested
        elif case == 2: extended_bit =     numpad_requested
        else:           extended_bit = True   # case 3
    except KeyError:
        extended_bit = False
    
    return virtual_key_code, extended_bit
    

## 
## Obtaining scan codes from virtual key codes:
## 

HKL  = HANDLE = PVOID = c_void_p
UINT = c_uint

GetKeyboardLayout = windll.user32.GetKeyboardLayout
GetKeyboardLayout.argtypes = [DWORD]
GetKeyboardLayout.restype  = HKL

MapVirtualKey = windll.user32.MapVirtualKeyW
MapVirtualKey.argtypes = [UINT, UINT]
MapVirtualKey.restype  = UINT

MapVirtualKeyEx = windll.user32.MapVirtualKeyExW
MapVirtualKeyEx.argtypes = [UINT, UINT, HKL]
MapVirtualKeyEx.restype  = UINT


def scan_code(virtual_key_code):
    # As far as I can tell, no need to use MapVirtualKeyEx here:
    
    #layout = GetKeyboardLayout(0)
    #return MapVirtualKeyEx(virtual_key_code, 0, layout)  # MAPVK_VK_TO_VSC = 0
    
    return MapVirtualKey(virtual_key_code, 0)  # MAPVK_VK_TO_VSC = 0
    

## 
## Atomic keyboard events:
## 

def virtual_key_event(generalized_key_code, releasing=False):
    virtual_key_code, extended_bit = unpack_generalized_key_code(generalized_key_code)
    flags = 0
    if releasing:     flags |= win32con.KEYEVENTF_KEYUP
    if extended_bit:  flags |= win32con.KEYEVENTF_EXTENDEDKEY
    # For many applications, a scan code of 0 seems to work fine and
    # might be faster:
    code = scan_code(virtual_key_code)
    return KeyboardInput(virtual_key_code, code, flags)


#
# A non-keyboard input method for sending a single Unicode character.
#
#     Ignores all modifier keys, is converted to nearest ASCII
# equivalent if destination window is an ASCII window.  These input
# events are relatively recent and some applications don't handle
# these events properly or at all (e.g., the Cygwin X server).
#
# char_code is a 16-bit Unicode code point (e.g., a single UCS-2 character)
#
def Unicode_event(char_code, releasing=False):
    flags = 0x0004  # KEYEVENTF_UNICODE
    if releasing: flags |= win32con.KEYEVENTF_KEYUP
    return KeyboardInput(0, char_code, flags)
    


### 
### Mouse events:
### 

# Warning: left here means the physical left mouse button, not the
# primary mouse button.  (These differ if the user has "Switch primary
# and secondary buttons" selected.)  Ditto for right.

Mouse_buttons = { 
    "left"  : [win32con.MOUSEEVENTF_LEFTDOWN,   win32con.MOUSEEVENTF_LEFTUP,   0],
    "right" : [win32con.MOUSEEVENTF_RIGHTDOWN,  win32con.MOUSEEVENTF_RIGHTUP,  0],
    "middle": [win32con.MOUSEEVENTF_MIDDLEDOWN, win32con.MOUSEEVENTF_MIDDLEUP, 0],
    # XBUTTON1 = 1
    "X1"    : [win32con.MOUSEEVENTF_XDOWN,      win32con.MOUSEEVENTF_XUP,      1],
    # XBUTTON2 = 2
    "X2"    : [win32con.MOUSEEVENTF_XDOWN,      win32con.MOUSEEVENTF_XUP,      2],
    }

def mouse_button_event(button, releasing=False):
    try:
        down_flag, up_flag, mouse_data = Mouse_buttons[button]
        flags = down_flag
        if releasing: flags = up_flag
        
        return MouseInput(0, 0, mouse_data, flags, 0)
    except KeyError:
        raise ValueError("unknown mouse button: " + button)


# clicks>0 => wheel rotated forward (if horizontal false) or else rotated right
def mouse_wheel_event(horizontal, clicks):
    flags = win32con.MOUSEEVENTF_WHEEL
    if horizontal: flags = 0x1000 # win32con.MOUSEEVENTF_HWHEEL
    amount = int(clicks * win32con.WHEEL_DELTA)
    return MouseInput(0, 0, amount, flags, 0)

# dx>0: moves right, dy>0: moves down
# absolute: 0..65535 each dim for primary monitor (virtual => entire desktop)
def mouse_move_event(x, y, absolute, virtual=False, coalesce=False):
    flags = win32con.MOUSEEVENTF_MOVE
    if not coalesce:
        flags |= win32con.MOUSEEVENTF_MOVE_NOCOALESCE   
    if absolute: 
        flags |= win32con.MOUSEEVENTF_ABSOLUTE
        if virtual:
            flags |= win32con.MOUSEEVENTF_VIRTUALDESK
    return MouseInput(x, y, 0, flags, 0)



### 
### Code for parsing extended SendDragonKeys syntax into a series of
### Input events suitable for calling SendInput with.
### 
### Uses ctypes (requires Python 2.5+).
### 
### Assumes input is (under Python 2) 8-bit Windows-1252 encoding or
### (under Python 3) Unicode encodable as Windows-1252.
### 
### Handling of control characters is at the mercy of the Windows
### VkKeyScan function; it sometimes makes odd choices like {pause}
### for 0x03.  Accordingly, I recommend avoiding these.  At least for
### the standard American keyboard, 0x0d (\r) presses {enter}.
### 
### Author:  Mark Lillibridge
### Version: 1.0
### 
#pylint:disable=C0321
from __future__ import print_function

import sys
import re
import win32con

from ctypes           import *
from dtactions.vocola_sendkeys.SendInput import *

debug = False

# ignore_unknown_names True means type out bad chords rather than
# raising a KeyError; e.g., "{bad}" sends {, b, a, d, }.
#
def senddragonkeys_to_events(input, ignore_unknown_names=True):
    chords = parse_into_chords(input)

    events = []
    for c in chords:
        try:
            events += chord_to_events(c)
        except LookupError as e:
            if not ignore_unknown_names: 
                raise
            if not c[0] and not c[2] and len(c[1])==1:
                raise  # already a single key chord
            characters = c[3]
            if debug:
                print("typing out bad chord: " + characters + ": " + repr(e))
            for char in characters:
                events += chord_to_events([None, char, None, char])

    return events

    

### 
### Break SendDragonKeys input into the chords that make it up.  Each
### chord is represented in terms of its three parts: modifiers, base,
### and effect.
### 
### E.g., "a{shift+left_10} " -> [[None, "a", None], ["shift", "left",
###                               "10"], [None, "space", None]]
### 
### Update: The chord's text is also stored for unparsing without information loss.
###         E.g., "{{}" -> [None, "{", None, "{{}"]
### 

def parse_into_chords(specification):
    chords = []
    
    while len(specification) > 0:
        m = chord_pattern.match(specification)
        if m:
            modifiers = m.group(1)
            if modifiers: modifiers = modifiers[:-1]  # remove final "+"
            chords += [[modifiers, m.group(2), m.group(3), m.group(0)]]
            specification = specification[m.end():]
        else:
            char = specification[0]
            chords += [[None, char, None, char]]
            specification = specification[1:]
    
    return chords

if sys.version_info[0] < 3:
    # Because we can't be sure of the current code page, treat all non-ASCII
    # characters as potential accented letters for now.  
    chord_pattern = re.compile(r"""\{ ( (?: [a-zA-Z0-9\x80-\xff]+ \+ )* ) 
                                      ( . | [-a-zA-Z0-9/*+.\x80-\xff]+ )
                                      (?: [ _] (\d+|hold|release) )?
                                   \}""", re.VERBOSE|re.IGNORECASE)
else:
    # for now, just assume modifiers and multiple-letter base names are pure ASCII:
    chord_pattern = re.compile(r"""\{ ( (?: [a-zA-Z0-9]+ \+ )* ) 
                                      ( . | [-a-zA-Z0-9/*+.]+ )
                                      (?: [ _] ([0-9]+|hold|release) )?
                                   \}""", re.VERBOSE|re.IGNORECASE)


### 
### 
### 

def chord_to_events(chord):
    modifiers, base, effect, text = chord
    if base == " ":
        base = "space"
    if modifiers:
        modifiers = modifiers.split("+")
    else:
        modifiers = []
    hold_count = release_count = 1
    if effect:
        effect = effect.lower()
        if   effect == "hold":    release_count = 0
        elif effect == "release": hold_count    = 0
        else: 
            hold_count = int(effect)
            if hold_count == 0:
                # check for bad names even when no events:
                for modifier in modifiers:
                    single(modifier, False)
                single(base, False)   
                return []

    if len(base) == 1:
        try:
            m, f = how_type_character(base)
            if debug and (len(m)>0 or describe_key(f)!=base):
                mm = ""
                if m: mm = '+'.join(m) + "+"
                bb = "<" + base + ">"
                if ord(base[0])<32: bb = hex(ord(base[0]))
                print("typing " + bb + " by {" + mm + describe_key(f) + "}")
            modifiers += m
            base = "VK" + hex(f)
        except:
            if debug and ord(base[0])<128:
                bb = "<" + base + ">"
                if ord(base[0])<32: bb = hex(ord(base[0]))
                print("can't type " + bb + " on current keyboard layout")
            pass
    
    events         = []
    modifiers_down = []
    modifiers_up   = []
    for modifier in modifiers:
        modifiers_down +=                single(modifier, False)
        modifiers_up   =  modifiers_up + single(modifier, True)

    try:
        # down down up (hardware auto-repeat style) fails so use down,up pairs:
        if hold_count > 1:
            return modifiers_down \
                + (single(base,False)+single(base, True))*hold_count \
                + modifiers_up
        if hold_count > 0:
            events += modifiers_down + single(base,False)*hold_count
        if release_count > 0:
            events += single(base, True) + modifiers_up
        return events
    except:
        if len(base) != 1:
            raise

    if len(modifiers) != 0:
        print("Warning: unable to use modifiers with character: " + base)
    
    # Unicode?
    
    if release_count==0:
        print("Warning: unable to independently hold character: " + base)
    if hold_count==0:
        print("Warning: unable to independently release character: " + base)
        return []

    windows_char = get_windows_1252_char(base)
    if debug:
        print("using numpad entry " + str(ord(windows_char)) + " for: " + base)
    return windows1252_to_events(ord(windows_char)) * hold_count



### 
### Pressing/releasing a single generalized virtual key or mouse button
### 

## 
## Keyboard key names:
## 
##   If you add names not matching [-a-zA-Z0-9/*+.]+, adjust chord_pattern.
## 

Key_name = {
    # 
    # SendDragonKeys virtual key names:
    # 
    
    "alt"         : VK_MENU,
    "backspace"   : VK_BACK,
    "break"       : VK_CANCEL,
    "capslock"    : VK_CAPITAL,
    "center"      : VK_CLEAR,
    "ctrl"        : VK_CONTROL,
    "del"         : VK_DELETE,
    "down"        : VK_DOWN,
    "end"         : VK_END,
    "enter"       : VK_RETURN,
    "esc"         : VK_ESCAPE,
    "home"        : VK_HOME,
    "ins"         : VK_INSERT,
    "left"        : VK_LEFT,
    "numlock"     : VK_NUMLOCK,
    "pgdn"        : VK_NEXT,
    "pgup"        : VK_PRIOR,
    "pause"       : VK_PAUSE,
    "prtsc"       : VK_SNAPSHOT,
    "right"       : VK_RIGHT,
    "scrolllock"  : VK_SCROLL,
    "shift"       : VK_SHIFT,
    "space"       : VK_SPACE,
    #"sysreq"     : VK_SYSREQ,# <<<>>>
    "tab"         : VK_TAB,
    "up"          : VK_UP,

    "f1"          : VK_F1,
    "f2"          : VK_F2,
    "f3"          : VK_F3,
    "f4"          : VK_F4,
    "f5"          : VK_F5,
    "f6"          : VK_F6,
    "f7"          : VK_F7,
    "f8"          : VK_F8,
    "f9"          : VK_F9,
    "f10"         : VK_F10,
    "f11"         : VK_F11,
    "f12"         : VK_F12,
    "f13"         : VK_F13,
    "f14"         : VK_F14,
    "f15"         : VK_F15,
    "f16"         : VK_F16,

    "numkey/"     : VK_DIVIDE,
    "numkey*"     : VK_MULTIPLY,
    "numkey-"     : VK_SUBTRACT,
    "numkey+"     : VK_ADD,
    "numkey0"     : VK_NUMPAD0,
    "numkey1"     : VK_NUMPAD1,
    "numkey2"     : VK_NUMPAD2,
    "numkey3"     : VK_NUMPAD3,
    "numkey4"     : VK_NUMPAD4,
    "numkey5"     : VK_NUMPAD5,
    "numkey6"     : VK_NUMPAD6,
    "numkey7"     : VK_NUMPAD7,
    "numkey8"     : VK_NUMPAD8,
    "numkey9"     : VK_NUMPAD9,
    "numkey."     : VK_DECIMAL,
    "numkeyenter" : GK_NUM_RETURN,

    "extdel"      : GK_NUM_DELETE,
    "extdown"     : GK_NUM_DOWN,
    "extend"      : GK_NUM_END,
    "exthome"     : GK_NUM_HOME,
    "extins"      : GK_NUM_INSERT,
    "extleft"     : GK_NUM_LEFT,
    "extpgdn"     : GK_NUM_NEXT,
    "extpgup"     : GK_NUM_PRIOR,
    "extright"    : GK_NUM_RIGHT,
    "extup"       : GK_NUM_UP,

    "leftalt"     : VK_LMENU,
    "rightalt"    : VK_RMENU,
    "leftctrl"    : VK_LCONTROL,
    "rightctrl"   : VK_RCONTROL,
    "leftshift"   : VK_LSHIFT,
    "rightshift"  : VK_RSHIFT,

    "0"           : VK_0,
    "1"           : VK_1,
    "2"           : VK_2,
    "3"           : VK_3,
    "4"           : VK_4,
    "5"           : VK_5,
    "6"           : VK_6,
    "7"           : VK_7,
    "8"           : VK_8,
    "9"           : VK_9,

    "a"           : VK_A,
    "b"           : VK_B,
    "c"           : VK_C,
    "d"           : VK_D,
    "e"           : VK_E,
    "f"           : VK_F,
    "g"           : VK_G,
    "h"           : VK_H,
    "i"           : VK_I,
    "j"           : VK_J,
    "k"           : VK_K,
    "l"           : VK_L,
    "m"           : VK_M,
    "n"           : VK_N,
    "o"           : VK_O,
    "p"           : VK_P,
    "q"           : VK_Q,
    "r"           : VK_R,
    "s"           : VK_S,
    "t"           : VK_T,
    "u"           : VK_U,
    "v"           : VK_V,
    "w"           : VK_W,
    "x"           : VK_X,
    "y"           : VK_Y,
    "z"           : VK_Z,


    # 
    # New names for virtual keys:
    # 

    "win"         : VK_LWIN,
    "leftwin"     : VK_LWIN,
    "rightwin"    : VK_RWIN,
    "apps"        : VK_APPS,  # name may change...

    "f17"         : VK_F17,
    "f18"         : VK_F18,
    "f19"         : VK_F19,
    "f20"         : VK_F20,
    "f21"         : VK_F21,
    "f22"         : VK_F22,
    "f23"         : VK_F23,
    "f24"         : VK_F24,

    "browserback"         : VK_BROWSER_BACK,
    "browserfavorites"    : VK_BROWSER_FAVORITES,
    "browserforward"      : VK_BROWSER_FORWARD,
    "browserhome"         : VK_BROWSER_HOME,
    "browserrefresh"      : VK_BROWSER_REFRESH,
    "browsersearch"       : VK_BROWSER_SEARCH,
    "browserstop"         : VK_BROWSER_STOP,

    # these names may change in the future...
    "launchapp1"          : VK_LAUNCH_APP1,
    "launchapp2"          : VK_LAUNCH_APP2,
    "launchmail"          : VK_LAUNCH_MAIL,
    "launchmediaselect"   : VK_LAUNCH_MEDIA_SELECT,

    "medianexttrack"      : VK_MEDIA_NEXT_TRACK,
    "mediaplaypause"      : VK_MEDIA_PLAY_PAUSE,
    "mediaprevioustrack"  : VK_MEDIA_PREV_TRACK,
    "mediastop"           : VK_MEDIA_STOP,

    "volumedown"          : VK_VOLUME_DOWN,
    "volumemute"          : VK_VOLUME_MUTE,
    "volumeup"            : VK_VOLUME_UP,

    # possibly more names to come...
    "oem1"      : VK_OEM_1,
    "oem2"      : VK_OEM_2,
    "oem3"      : VK_OEM_3,
    "oem4"      : VK_OEM_4,
    "oem5"      : VK_OEM_5,
    "oem6"      : VK_OEM_6,
    "oem7"      : VK_OEM_7,
    "oem8"      : VK_OEM_8,
    "oem102"    : VK_OEM_102,
    "oemcomma"  : VK_OEM_COMMA,
    "oemminus"  : VK_OEM_MINUS,
    "oemperiod" : VK_OEM_PERIOD,
    "oemplus"   : VK_OEM_PLUS,
}

Code_to_name = {}
for name in list(Key_name.keys()):
    Code_to_name[Key_name[name]] = name

def describe_key(code):
    try:
        return Code_to_name[code]
    except:
        return "VK" + hex(code)


## 
## Mouse button names:
## 

Button_name = {
    "leftbutton"   : "left",    # really primary button
    "middlebutton" : "middle",
    "rightbutton"  : "right",   # really secondary button
    "xbutton1"     : "X1",
    "xbutton2"     : "X2",
    }


GetSystemMetrics = windll.user32.GetSystemMetrics
GetSystemMetrics.argtypes = [c_int]
GetSystemMetrics.restype  = c_int

# Convert vocola_ExtendSendDragonKeys mouse button names to those
# required by vocola_SendInput.py, swapping left & right buttons if
# user has "Switch primary and secondary buttons" selected:
def get_mouse_button(button_name):
    try:
        button = Button_name[button_name.lower()]
        if button=="left" or button=="right":
            if GetSystemMetrics(win32con.SM_SWAPBUTTON):
                if button=="left":
                    button = "right"
                else:
                    button = "left"
        return button
    except:
        raise KeyError("unknown mouse button: " + key)


## 
## Create a single virtual event to press or release a keyboard key or
## mouse button:
## 

def single(key, releasing):
    # universal syntax is VK0xhh for virtual key with code 0xhh:
    if key[0:4] == "VK0x":
        return [virtual_key_event(int(key[4:],16), releasing)]
        
    lower_key = key.lower()
    try:
        return [virtual_key_event(Key_name[lower_key], releasing)]
    except:
        try:
            return [mouse_button_event(get_mouse_button(lower_key), releasing)]
        except:
            raise KeyError("unknown key/button: " + key)



### 
### 
### 

DWORD  = c_ulong        # 32 bits
SHORT  = c_short        # 16 bits
#TCHAR = c_char         # if not using Unicode
TCHAR  = c_wchar        # if using Unicode
HKL    = HANDLE = PVOID = c_void_p

GetKeyboardLayout = windll.user32.GetKeyboardLayout
GetKeyboardLayout.argtypes = [DWORD]
GetKeyboardLayout.restype  = HKL

VkKeyScan = windll.user32.VkKeyScanW
VkKeyScan.argtypes = [TCHAR]
VkKeyScan.restype  = SHORT

VkKeyScanEx = windll.user32.VkKeyScanExW
VkKeyScanEx.argtypes = [TCHAR, HKL]
VkKeyScanEx.restype  = SHORT


def how_type_character(char):
    how_type = VkKeyScan(char)
    
    virtual_key = how_type & 0xff
    if virtual_key == 0xff:
        raise ValueError("unable to type character with current keyboard layout: " 
                         + char)

    modifiers   = []
    if how_type&0x400: modifiers += ["alt"]
    if how_type&0x200: modifiers += ["ctrl"]
    if how_type&0x100: modifiers += ["shift"]

    if how_type&0xf800:
        raise ValueError("unknown modifiers required, tell MDL: " + hex(how_type)) 
    
    return modifiers, virtual_key


### 
### 
### 

def get_windows_1252_char(char):
    if sys.version_info[0] > 2:
        try:
            return char.encode('Windows-1252')
        except:
            raise LookupError("SendInput passed non-Windows-1252 character: " + char)
    return char


def windows1252_to_events(code):
    events = []
    events += single("alt", False)
    events += numpad(0)
    events += numpad(code//100 %10)
    events += numpad(code//10  %10)
    events += numpad(code//1   %10)
    events += single("alt", True)
    return events

def numpad(i):
    return chord_to_events([None, "numkey"+str(i), None, "{numkey"+str(i)+"}"])



### 
### 
### 

Wheel_name = {
    "wheelup"    : [1, False],
    "wheeldown"  : [-1, False],
    "wheelright" : [1, True],
    "wheelleft"  : [-1, True]
    }
